/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <cassert>
#include <gtk/gtk.h>
#include <pan/general/debug.h>
#include <pan/general/e-util.h>
#include <pan/general/foreach.h>
#include <pan/general/quark.h>
#include <pan/general/i18n.h>
#include <pan/general/log.h>
#include <pan/usenet-utils/gnksa.h>
#include <pan/data/article.h>
#include <pan/data/data.h>
#include <pan/data/filter-info.h>
#include <pan/icons/pan-pixbufs.h>
#include "header-pane.h"
#include "pad.h"

using namespace pan;

/****
*****
****/

void
HeaderPane :: ArticleViewRec :: set_article (const Article * a)
{
  article = a;
  short_author = GNKSA :: get_short_author_name (article->author.to_view());
}

const char*
HeaderPane :: ArticleViewRec :: get_subject ()
{
  if (collated_subject == 0) // lazy instantiation
  {
    // casefolded utf-8-friendly comparison key...
    StringView original (article->subject.to_view());
    while (!original.empty() && !isalpha(*original.str))
      original.eat_chars (1);
    char * casefold = g_utf8_casefold (original.str, original.len);
    collated_subject = g_utf8_collate_key (casefold, -1);
    g_free (casefold);
  }

  return collated_subject;
}

const char*
HeaderPane :: ArticleViewRec :: get_author ()
{
  if (collated_author == 0) // lazy instantiation
  {
    // casefolded utf-8-friendly comparison key...
    char * casefold = g_utf8_casefold (short_author.str, short_author.len);
    collated_author = g_utf8_collate_key (casefold, -1);
    g_free (casefold);
  }

  return collated_author;
}

HeaderPane :: ArticleViewRec :: ~ArticleViewRec ()
{
  g_free (collated_subject);
  g_free (collated_author);
}

namespace
{
  GQuark pane_quark () {
    static GQuark q (0);
    if (!q) q = g_quark_from_static_string ("pane");
    return q;
  }

  GQuark view_quark () {
    static GQuark q (0);
    if (!q) q = g_quark_from_static_string ("view");
    return q;
  }

  HeaderPane::ArticleViewRec& get_rec (GtkTreeModel      * model,
                                       GtkTreeIter       * iter)
  {
    HeaderPane * pane (static_cast<HeaderPane*>(g_object_get_qdata (G_OBJECT(model), pane_quark())));
    return pane->_iter_to_view_record[iter->user_data];
  }
}

/****
*****
****/

namespace
{
  typedef std::set<const Article*> articles_t;

  struct Icon
  {
    const guint8 * pixbuf_txt;
    GdkPixbuf * pixbuf;
  };

  enum
  {
    ICON_READ,
    ICON_UNREAD,
    ICON_NEW,
    ICON_COMPLETE,
    ICON_COMPLETE_READ,
    ICON_INCOMPLETE,
    ICON_INCOMPLETE_READ,
    ICON_CACHED,
    //ICON_QUEUED,
    ICON_ERROR,
    //ICON_FLAGGED_FOR_DL,
    ICON_QTY
   };

   Icon _icons[ICON_QTY] =
   {
      { icon_article_read,      0 },
      { icon_article_unread,    0 },
      { icon_article_new,       0 },
      { icon_binary_complete,   0 },
      { icon_binary_complete_read,   0 },
      { icon_binary_incomplete, 0 },
      { icon_binary_incomplete_read, 0 },
      { icon_disk,              0 },
     // { icon_bluecheck,         0 },
      { icon_x,                 0 }
      //{ icon_blue_flag,         0 }
   };

   const int
   get_article_state (const Data& data, const Article * a)
   {
      int retval;
      const bool read (data.is_read (a));
      const int part_state (a->get_part_state());

      if (!a /* || a->error_flag*/)
         retval = ICON_ERROR;
      else if (part_state==Article::COMPLETE && read)
         retval = ICON_COMPLETE_READ;
      else if (part_state==Article::COMPLETE)
         retval = ICON_COMPLETE;
      else if (part_state==Article::INCOMPLETE && read)
         retval = ICON_INCOMPLETE_READ;
      else if (part_state==Article::INCOMPLETE)
         retval = ICON_INCOMPLETE;
      else if (read)
         retval = ICON_READ;
      else
         retval = ICON_UNREAD;
                                                                                                                                              
      return retval;
   }

  const Icon*
  get_column_state_icon (const Data& data, const Article * a)
  {
    return _icons + get_article_state (data, a);
  }

  const Icon*
  get_column_action_icon (const Article * a, const ArticleCache& cache)
  {
    const Icon * retval (0);
      
#if 0
    if (flagset_has_article (a))
      retval = _icons + ICON_FLAGGED_FOR_DL;
    else if (queue_get_message_id_status (&a->message_id))
      retval = _icons + ICON_QUEUED;
    else
#endif
    if (cache.contains (a->message_id))
      retval = _icons + ICON_CACHED;

    return retval;
  }
};

namespace
{
  enum
  {
    COL_DATE_STR,
    COL_STATE,
    COL_ACTION,
    COL_SUBJECT,
    COL_SCORE,
    COL_LINES,
    COL_AUTHOR,
    COL_DATE,
    COL_QTY
  };

  void
  render_action (GtkTreeViewColumn * col,
                 GtkCellRenderer   * renderer,
                 GtkTreeModel      * model,
                 GtkTreeIter       * iter,
                 gpointer            user_data)
  {
    const Article *a (get_rec (model, iter).article);
    const HeaderPane * h = (const HeaderPane*) user_data;
    const Icon * icon = get_column_action_icon (a, h->_cache);
    g_object_set (renderer, "pixbuf", icon ? icon->pixbuf : NULL, NULL);
  }

  void
  render_state (GtkTreeViewColumn * col,
                GtkCellRenderer   * renderer,
                GtkTreeModel      * model,
                GtkTreeIter       * iter,
                gpointer            user_data)
  {
    int state (0);
    gtk_tree_model_get (model, iter, COL_STATE, &state, -1);
    g_object_set (renderer, "pixbuf", _icons[state].pixbuf, NULL);
  }

  int
  find_highest_followup_score (GtkTreeModel * model,
                               GtkTreeIter * parent)
  {
    int score (-9999);
    GtkTreeIter child;
    if (gtk_tree_model_iter_children (model, &child, parent)) do {
      const Article * a (get_rec(model, &child).article);
      score = std::max (score, a->score);
      score = std::max (score, find_highest_followup_score (model, &child));
    } while (gtk_tree_model_iter_next (model, &child));
    return score;
  }

  void
  render_score (GtkTreeViewColumn * col,
                GtkCellRenderer   * renderer,
                GtkTreeModel      * model,
                GtkTreeIter       * iter,
                gpointer            user_data)
  {
    const Article * article (get_rec(model, iter).article);
    int score (article->score);
    char buf[16];
    if (score)
      g_snprintf(buf, sizeof(buf), "%d", score);
    else
      *buf = '\0';

    GtkWidget * view ((GtkWidget*) g_object_get_qdata (G_OBJECT(model), view_quark()));
    GtkTreePath * path (gtk_tree_model_get_path (model, iter));
    const bool expanded (gtk_tree_view_row_expanded (GTK_TREE_VIEW(view), path));
    gtk_tree_path_free (path);
    if (!expanded)
      score = std::max (score, find_highest_followup_score (model, iter));

    const Prefs& prefs (*static_cast<const Prefs*>(user_data));
    std::string bg, fg;
    if (score >= 9999) {
      fg = prefs.get_color_str ("score-color-watched-fg", "black");
      bg = prefs.get_color_str ("score-color-watched-bg", "green");
    } else if (score >= 5000) {
      fg = prefs.get_color_str ("score-color-high-fg", "black");
      bg = prefs.get_color_str ("score-color-high-bg", "yellow");
    } else if (score >= 1) {
      fg = prefs.get_color_str ("score-color-medium-fg", "black");
      bg = prefs.get_color_str ("score-color-medium-bg", "red");
    } else if (score <= -9999) {
      fg = prefs.get_color_str ("score-color-ignored-fg", "dark grey");
      bg = prefs.get_color_str ("score-color-ignored-bg", "black");
    } else if (score <= -5000) {
      fg = prefs.get_color_str ("score-color-low-fg", "light grey");
      bg = prefs.get_color_str ("score-color-low-bg", "black");
    }

    g_object_set (renderer, "text", buf,
                            "xalign", 1.0,
                            "weight", PANGO_WEIGHT_NORMAL,
                            "background", (bg.empty() ? NULL : bg.c_str()),
                            "foreground", (fg.empty() ? NULL : fg.c_str()),
                            NULL);
  }

  void
  render_author (GtkTreeViewColumn * col,
                 GtkCellRenderer   * renderer,
                 GtkTreeModel      * model,
                 GtkTreeIter       * iter,
                 gpointer            user_data)
  {
    HeaderPane::ArticleViewRec& rec (get_rec (model, iter));
    char * pch = g_strndup (rec.short_author.str, rec.short_author.len);
    g_object_set (renderer, "text", pch,
                            "xalign", 0.0,
                            "weight", PANGO_WEIGHT_NORMAL,
                            NULL);
    g_free (pch);
  }

  void
  render_lines (GtkTreeViewColumn * col,
                GtkCellRenderer   * renderer,
                GtkTreeModel      * model,
                GtkTreeIter       * iter,
                gpointer            user_data)
  {
    const Article * a (get_rec(model, iter).article);
    const unsigned long lines (a->get_line_count ());
    char buf[32];
    g_snprintf (buf, sizeof(buf), "%lu", lines);
    g_object_set (renderer, "text", buf,
                            "xalign", 1.0,
                            "weight", PANGO_WEIGHT_NORMAL,
                            NULL);
  }

  bool
  has_unread_children (GtkTreeModel * model,
                       GtkTreeIter * parent,
                       const Data * data)
  {
    GtkTreeIter child;
    if (gtk_tree_model_iter_children (model, &child, parent)) do {
      const Article * a (get_rec(model, &child).article);
      if (a && !data->is_read (a))
        return true;
      if (has_unread_children (model, &child, data))
        return true;
    } while (gtk_tree_model_iter_next (model, &child));

    return false;
  }

  void
  render_subject (GtkTreeViewColumn * col,
                  GtkCellRenderer   * renderer,
                  GtkTreeModel      * model,
                  GtkTreeIter       * iter,
                  gpointer            data_g)
  {
    const Data * data (static_cast<const Data*>(data_g));
    const Article *a (get_rec (model, iter).article);
    const bool bold (!data->is_read(a));

    bool underlined (false);
    if (data->is_read(a)) {
      GtkWidget * view ((GtkWidget*) g_object_get_qdata (G_OBJECT(model), view_quark()));
      GtkTreePath * path (gtk_tree_model_get_path (model, iter));
      const bool expanded (gtk_tree_view_row_expanded (GTK_TREE_VIEW(view), path));
      gtk_tree_path_free (path);
      underlined = !expanded && has_unread_children (model, iter, data);
    }

    g_object_set (renderer,
      "text", a->subject.c_str(),
      "xalign", 0.0,
      "weight", (bold ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL),
      "underline", (underlined ? PANGO_UNDERLINE_SINGLE : PANGO_UNDERLINE_NONE),
      NULL);
  }
}

namespace
{
  typedef std::set<GtkTreePath*> paths_t;

  void
  add_article_to_model (const Data                 & data,
                        PanTreeStore               * store, 
                        GtkTreeIter                * parent_iter,
                        const Article              * article,
                        GtkTreeIter                * setme_child_iter=0)
  {
    // insert the row...
    GtkTreeIter child_iter;
    pan_tree_store_append (store, &child_iter, parent_iter);

    // populate the row...
    char * date_str (get_evolution_style_date_string (article->time_posted));

    // GtkTreeStore's iters only use the first user_data field,
    // and since it's a GNode it won't change unless the row
    // is removed... so its pointer is a safe place to hang our keys
    // so long as we remember to erase elements in pan_tree_store_remove()
    get_rec (GTK_TREE_MODEL(store), &child_iter).set_article (article);

    pan_tree_store_set (store, &child_iter,
                        COL_STATE, get_article_state(data, article),
                        COL_DATE_STR, date_str,
                        -1);
    g_free (date_str);

    if (setme_child_iter)
      *setme_child_iter = child_iter;
  }

  void
  add_children_to_model (const Data                 & data,
                         PanTreeStore               * store,
                         GtkTreeIter                * parent_iter,
                         const Quark                & parent_mid,
                         const Data::ArticleTree    * atree)
  {
    typedef Data::ArticleTree::articles_t articles_t; 
    articles_t children;
    atree->get_children (parent_mid, children);

    foreach_const (articles_t, children, it)
    {
      GtkTreeIter child_iter;
      add_article_to_model (data, store, parent_iter, *it, &child_iter);
      add_children_to_model (data, store, &child_iter, (*it)->message_id, atree);
    }
  }

  int
  column_compare_func (GtkTreeModel  * model,
                       GtkTreeIter   * iter_a,
                       GtkTreeIter   * iter_b,
                       gpointer        userdata)
  {
    int ret (0);
    int sortcol (GPOINTER_TO_INT (userdata));
    HeaderPane * pane (static_cast<HeaderPane*>(g_object_get_qdata (G_OBJECT(model), pane_quark())));

    GtkTreeIter unused;
    const bool is_toplevel (!gtk_tree_model_iter_parent (model, &unused, iter_a));
    if (!is_toplevel)
      sortcol = COL_DATE;

    switch (sortcol)
    {
      case COL_STATE: {
        int a(0), b(0);
        gtk_tree_model_get (model, iter_a, COL_STATE, &a, -1);
        gtk_tree_model_get (model, iter_b, COL_STATE, &b, -1);
        ret = a - b;
        break;
      }

      case COL_ACTION: {
        int action_a(0), action_b(0);
        gtk_tree_model_get (model, iter_a, COL_ACTION, &action_a, -1);
        gtk_tree_model_get (model, iter_b, COL_ACTION, &action_b, -1);
        ret = action_a - action_b;
        break;
      }

      case COL_SUBJECT: {
        ret = strcmp (pane->_iter_to_view_record[iter_a->user_data].get_subject(),
                      pane->_iter_to_view_record[iter_b->user_data].get_subject());
        break;
      }

      case COL_SCORE: {
        const Article *a (pane->_iter_to_view_record[iter_a->user_data].article);
        const Article *b (pane->_iter_to_view_record[iter_b->user_data].article);
        ret = a->score - b->score;
        break;
      }

      case COL_LINES: {
        const Article *a (pane->_iter_to_view_record[iter_a->user_data].article);
        const Article *b (pane->_iter_to_view_record[iter_b->user_data].article);
        const unsigned long a_lines (a->get_line_count ());
        const unsigned long b_lines (b->get_line_count ());
        if (a_lines < b_lines)
           ret = -1;
        else if (a_lines > b_lines)
           ret = 1;
        else
           ret = 0;
        break;
      }

      case COL_AUTHOR: {
        ret = strcmp (pane->_iter_to_view_record[iter_a->user_data].get_author(),
                      pane->_iter_to_view_record[iter_b->user_data].get_author());
        break;
      }

      case COL_DATE: {
        const Article *a (pane->_iter_to_view_record[iter_a->user_data].article);
        const Article *b (pane->_iter_to_view_record[iter_b->user_data].article);
        if (a->time_posted < b->time_posted)
          ret = -1;
        else if (a->time_posted > b->time_posted)
          ret = 1;
        else
          ret = 0;
        break;
      }

      default:
        assert (0);
    }

#if 0
    // this is an in-place sort
    if (!ret) {
       GtkTreePath * p1 = gtk_tree_model_get_path (model, iter_a);
       GtkTreePath * p2 = gtk_tree_model_get_path (model, iter_b);
       if (!p1 && !p2) ret = 0;
       else if (!p1)   ret = -1;
       else if (!p2)   ret = 1;
       else            ret = gtk_tree_path_compare (p1, p2);
       gtk_tree_path_free (p2);
       gtk_tree_path_free (p1);
    }
#endif

    return ret;
  }

  PanTreeStore*
  build_model (HeaderPane                * pane,
               const Data                & data,
               const Quark               & group,
               Data::ArticleTree         * atree,
               const TextMatch           * match)
  {
    // build the model...
    // only 'state', 'action', and 'date_string' are used;
    // the rest are just needed to give ccc placeholders for sorting,
    // so they've been given the smallest type available...
    PanTreeStore * store = pan_tree_store_new (COL_QTY,
                                               G_TYPE_STRING, // date string
                                               G_TYPE_INT, // state
                                               G_TYPE_INT, // action
                                               G_TYPE_CHAR, // subject
                                               G_TYPE_CHAR, // score
                                               G_TYPE_CHAR, // lines
                                               G_TYPE_CHAR, // author
                                               G_TYPE_CHAR); //  date
    g_object_set_qdata (G_OBJECT(store), pane_quark(), pane);

    GtkTreeSortable * sort = GTK_TREE_SORTABLE(store);
    for (int i=0; i<COL_QTY; ++i)
      gtk_tree_sortable_set_sort_func (sort, i, column_compare_func, GINT_TO_POINTER(i), NULL);
    if (!group.empty())
       add_children_to_model (data, store, NULL, Quark(), atree);

    return store;
  }
}

bool
HeaderPane :: set_group (const Quark& new_group)
{
  const Quark old_group (_group);
  bool change (old_group != new_group);

  if (change)
  {
    _group = new_group;

    if (_atree)
    {
      _atree->remove_listener (this);
      delete _atree;
      _atree = 0;
    }

    if (!new_group.empty())
    {
      _atree = _data.group_get_articles (new_group, &_filter);
      _atree->add_listener (this);

      _tree_store = build_model (this, _data, _group, _atree, NULL);
      gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(_tree_store),
                                            COL_DATE, GTK_SORT_DESCENDING);
      g_object_set_qdata (G_OBJECT(_tree_store), view_quark(), _tree_view);
      gtk_tree_view_set_model (GTK_TREE_VIEW(_tree_view),
                               GTK_TREE_MODEL(_tree_store));
      g_object_unref (G_OBJECT(_tree_store)); // store is deleted w/view
    }
  }

  return change;
}

/****
*****
****/

namespace
{
  struct RefreshStateForeachData
  {
    const Data * _data;
    const quarks_t * _refreshme;
    RefreshStateForeachData (const Data * data, const quarks_t * refreshme=0):
      _data(data), _refreshme(refreshme) {}
  };

  gboolean
  refresh_state_col_foreach (GtkTreeModel * model,
                             GtkTreePath * path,
                             GtkTreeIter * iter,
                             gpointer r_gpointer)
  {
    const Article *a (get_rec (model, iter).article);
    int old_state;
    gtk_tree_model_get (model, iter, COL_STATE, &old_state, -1);
    const RefreshStateForeachData * r (static_cast<const RefreshStateForeachData*>(r_gpointer));
    if (!r->_refreshme || r->_refreshme->count(a->message_id)) {
      const int new_state (get_article_state (*r->_data, a));
      if (new_state != old_state) {
        GtkTreeIter i = *iter;
        pan_tree_store_set (PAN_TREE_STORE(model), &i, COL_STATE, new_state, -1);
      }
    }

    return false;
  }

  void
  refresh_state_col (const Data * data, GtkTreeModel * model, const quarks_t * refreshme=0)
  {
    RefreshStateForeachData tmp (data, refreshme);
    gtk_tree_model_foreach (model, refresh_state_col_foreach, &tmp);
  }
}

void
HeaderPane :: on_group_read (const Quark& group)
{
  if (group == _group)
  {
    refresh_state_col (&_data, GTK_TREE_MODEL(_tree_store));
    gtk_widget_queue_draw (_tree_view);
  }
}

/****
*****
****/

namespace
{
  typedef std::map<Quark,quarks_t> parent_to_children_t;

  void
  add_subtree_to_map (const Data::ArticleTree  & tree,
                      const Quark              & parent_mid, 
                      parent_to_children_t     & setme)
  {
    Data::ArticleTree::articles_t children;
    tree.get_children (parent_mid, children);
    foreach_const (Data::ArticleTree::articles_t, children, it)
    {
      const Quark child_mid ((*it)->message_id);
      setme[parent_mid].insert (child_mid);
      add_subtree_to_map (tree, child_mid, setme);
    }
  }

  void
  walk_and_remove (GtkTreeModel     * model,
                   GtkTreeIter      * cur,
                   const quarks_t   & mids_to_remove)
  {
    for (;;)
    {
      const Article *a (get_rec (model, cur).article);

      // process the children first
      GtkTreeIter child;
      if (gtk_tree_model_iter_children (model, &child, cur))
        walk_and_remove (model, &child, mids_to_remove);

      // if this article's not being removed, proceed to its sibling
      if (!mids_to_remove.count (a->message_id))
      {
        if (!gtk_tree_model_iter_next (model, cur))
          break; // that was the last sibling
      }
      else // this article's going away
      {
        // remove it from our iter-to-article map...
        HeaderPane * pane (static_cast<HeaderPane*>(g_object_get_qdata (G_OBJECT(model), pane_quark())));
        pane->_iter_to_view_record.erase (cur->user_data);
        if (!pan_tree_store_remove (PAN_TREE_STORE(model), cur))
          break; // that was the last sibling
      }
    }
  }

  void
  walk_and_add (const Data                  & data,
                GtkTreeModel                * model,
                GtkTreeIter                 * parent,
                GtkTreeIter                 * cur,
                const parent_to_children_t  & mids_with_new_children,
                const Data::ArticleTree     * atree)
  {
    for (;;)
    {
      // find the Article associated with this TreeNode
      const Article *a (get_rec (model, cur).article);

      // add all of this Article's new children to the model
      parent_to_children_t::const_iterator i (mids_with_new_children.find (a->message_id));
      if (i != mids_with_new_children.end())
        foreach_const (quarks_t, i->second, it)
          add_article_to_model (data, PAN_TREE_STORE(model), cur, atree->get_article(*it));

      // recurse to the children and repeat
      GtkTreeIter child;
      if (gtk_tree_model_iter_children (model, &child, cur))
        walk_and_add (data, model, cur, &child, mids_with_new_children, atree);

      // do the same for the siblings
      if (!gtk_tree_model_iter_next (model, cur))
        break;
    }
  }
}

void
HeaderPane :: on_tree_change (const Data::ArticleTree::Diffs& diffs)
{
  debug (diffs.added.size() << " article added; "
      << diffs.reparented.size() << " reparented; "
      << diffs.removed.size() << " removed");

  GtkTreeModel * model = GTK_TREE_MODEL(_tree_store);

  const bool added_or_reparented (!diffs.added.empty() ||
                                  !diffs.reparented.empty());
                  //!diffs.removed.empty());

  // for speed's sake, disable sorting during the make batch changes
  const GtkSortType order (PAN_TREE_STORE(model)->order);
  const int sort_column_id (PAN_TREE_STORE(model)->sort_column_id);
  if (added_or_reparented)
    gtk_tree_sortable_set_sort_column_id (
      GTK_TREE_SORTABLE(model), 
      GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID,
      order);

  // build a list of subtrees to prune from diffs.removed + diffs.reparented.
  // treestore can't reparent, so reparenting is done by removing and re-adding.
  quarks_t mids_to_remove;
  mids_to_remove = diffs.removed;
  foreach_const (Data::ArticleTree::Diffs::reparented_t, diffs.reparented, it)
    mids_to_remove.insert (it->message_id);

  // build a list of nodes to add from diffs.added + diffs.reparented.
  // treestore can't reparent, so reparenting is done by removing and re-adding.
  quarks_t roots_to_add;
  parent_to_children_t nodes_with_new_children;
  foreach_const (Data::ArticleTree::Diffs::reparented_t, diffs.reparented, it) {
    const Quark& mid (it->message_id);
    if (it->new_parent.empty())
      roots_to_add.insert (mid);
    else
      nodes_with_new_children[it->new_parent].insert (mid);
    add_subtree_to_map (*_atree, mid, nodes_with_new_children);
  }
  foreach_const (Data::ArticleTree::Diffs::added_t, diffs.added, it) {
    const Quark& mid (it->message_id);
    if (it->parent.empty())
      roots_to_add.insert (mid);
    else
      nodes_with_new_children[it->parent].insert (mid);
    add_subtree_to_map (*_atree, mid, nodes_with_new_children);
  }

  debug ("mids_to_remove.size() " << mids_to_remove.size());
  debug ("roots_to_add.size() " << roots_to_add.size());
  debug ("nodes with new children size() " << nodes_with_new_children.size());

  GtkTreeIter iter;

  // remove some nodes.
  if (!mids_to_remove.empty())
    if (gtk_tree_model_get_iter_first (model, &iter))
      walk_and_remove (GTK_TREE_MODEL(_tree_store), &iter, mids_to_remove);

  foreach_const (quarks_t, roots_to_add, it)
    add_article_to_model (_data, _tree_store, 0, _atree->get_article(*it));

  if (!nodes_with_new_children.empty())
    if (gtk_tree_model_get_iter_first (model, &iter))
      walk_and_add (_data, GTK_TREE_MODEL(_tree_store), 0, &iter, nodes_with_new_children, _atree);


  // and turn sorting back on...
  if (added_or_reparented)
    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(model),
                                          sort_column_id, order);

  if (!diffs.changed.empty()) {
    refresh_state_col (&_data, GTK_TREE_MODEL(_tree_store), &diffs.changed);
    gtk_widget_queue_draw (GTK_WIDGET(_tree_view));
  }
}

/****
*****  SELECTION
****/

const Article*
HeaderPane :: get_first_selected_article () const
{
   const Article * a (0);
   std::set<const Article*> articles = get_full_selection ();
   if (!articles.empty())
     a = *articles.begin ();
   return a;
}

std::set<const Article*>
HeaderPane :: get_full_selection () const
{
  debug ("get_full_selection");
  std::set<const Article*> articles;

  // get a list of paths
  GtkTreeModel * model (0);
  GtkTreeSelection * selection (gtk_tree_view_get_selection (GTK_TREE_VIEW(_tree_view)));
  GList * list (gtk_tree_selection_get_selected_rows (selection, &model));
  for (GList *l(list); l; l=l->next) {
     GtkTreePath * path (static_cast<GtkTreePath*>(l->data));
     GtkTreeIter iter;
     if (gtk_tree_model_get_iter (model, &iter, path)) {
       const Article *a (get_rec (model, &iter).article);
       articles.insert (a);
     }
  }

  // cleanup
  g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (list);

  return articles;
}

namespace
{
  void
  walk_and_collect (GtkTreeModel                * model,
                    GtkTreeIter                 * cur,
                    articles_t                  & setme)
  {
    for (;;) {
      setme.insert (get_rec (model, cur).article);
      GtkTreeIter child;
      if (gtk_tree_model_iter_children (model, &child, cur))
        walk_and_collect (model, &child, setme);
      if (!gtk_tree_model_iter_next (model, cur))
        break;
    }
  }

  struct NestedData {
    const HeaderPane * pane;
    articles_t articles;
  };
}

void
HeaderPane :: get_nested_foreach (GtkTreeModel* model,
                                  GtkTreePath* path,
                                  GtkTreeIter* iter,
                                  gpointer data) const
{
  articles_t& articles (*static_cast<articles_t*>(data));
  articles.insert (get_rec (model, iter).article);
  const bool expanded (gtk_tree_view_row_expanded (GTK_TREE_VIEW(_tree_view), path));
  GtkTreeIter child;
  if (!expanded && gtk_tree_model_iter_children(model,&child,iter))
    walk_and_collect (model, &child, articles);
}
void
HeaderPane :: get_nested_foreach_static (GtkTreeModel* model,
                                         GtkTreePath* path,
                                         GtkTreeIter* iter,
                                         gpointer data)
{
  NestedData& ndata (*static_cast<NestedData*>(data));
  ndata.pane->get_nested_foreach (model, path, iter, &ndata.articles);
}

articles_t
HeaderPane :: get_nested_selection () const
{
  NestedData data;
  data.pane = this;
  GtkTreeSelection * selection (gtk_tree_view_get_selection (GTK_TREE_VIEW(_tree_view)));
  gtk_tree_selection_selected_foreach (selection, get_nested_foreach_static, &data);
  return data.articles;
}

/****
*****  POPUP MENU
****/

void
HeaderPane ::  do_popup_menu (GtkWidget *treeview, GdkEventButton *event, gpointer pane_g)
{
  HeaderPane * self (static_cast<HeaderPane*>(pane_g));
  GtkWidget * menu (self->_action_manager.get_action_widget ("/header-pane-popup"));
  gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
                  (event ? event->button : 0),
                  (event ? event->time : 0));
}

namespace
{
  gboolean on_popup_menu (GtkWidget * treeview, gpointer userdata)
  {
    HeaderPane::do_popup_menu (treeview, NULL, userdata);
    return true;
  }
}


gboolean
HeaderPane :: on_button_pressed (GtkWidget * treeview, GdkEventButton *event, gpointer userdata)
{
  HeaderPane * self (static_cast<HeaderPane*>(userdata));

  // single click with the right mouse button?
  if (event->type == GDK_BUTTON_PRESS  &&  event->button == 3)
  {
    GtkTreeSelection * selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
    GtkTreePath *path;
    if (gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW(treeview),
                                       (gint) event->x, 
                                       (gint) event->y,
                                       &path, NULL, NULL, NULL))
    {
      if (!gtk_tree_selection_path_is_selected (selection, path))
      {
        gtk_tree_selection_unselect_all (selection);
        gtk_tree_selection_select_path (selection, path);
      }
      gtk_tree_path_free(path);
    }
    HeaderPane::do_popup_menu(treeview, event, userdata);
    return true;
  }
  else if (self->_prefs.get_flag("single-click-activates-article",true)
           && (event->type == GDK_BUTTON_RELEASE)
           && (event->button == 1)
           && (event->window == gtk_tree_view_get_bin_window (GTK_TREE_VIEW(treeview)))
           && !(event->state & (GDK_SHIFT_MASK|GDK_CONTROL_MASK|GDK_MOD1_MASK)))
  {
    GtkTreePath * path;
    GtkTreeViewColumn * col;
    gint cell_x(0), cell_y(0);
    if (gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(treeview),
                                      (gint) event->x,
                                      (gint) event->y,
                                      &path, &col, &cell_x, &cell_y))
    {
      // get the cell area to make sure they didn't click on an expander.
  //    GdkRectangle rect;
  //    gtk_tree_view_get_cell_area (GTK_TREE_VIEW(treeview), path, col, &rect);
  //    if (rect.x<=cell_x && cell_x<(rect.x+rect.width) && rect.y<=cell_y && cell_y<(rect.y+rect.height))
        gtk_tree_view_row_activated (GTK_TREE_VIEW(treeview), path, col);
      gtk_tree_path_free(path);
    }
  }

  return false;
}

namespace
{
  bool is_probably_an_image (const Article& a)
  {
    const StringView s (a.subject.to_view());
    return s.strstr(".jpg") || s.strstr(".JPG") ||
           s.strstr(".gif") || s.strstr(".GIF") ||
           s.strstr(".jpeg") || s.strstr(".JPEG") ||
           s.strstr(".png") || s.strstr(".PNG");
  }

  gboolean on_row_activated_idle (gpointer pane_g)
  {
    HeaderPane * pane (static_cast<HeaderPane*>(pane_g));
    const Article * a (pane->get_first_selected_article());
    if (a) {
      if (a->get_line_count()<5000 || is_probably_an_image(*a))
        pane->_action_manager.activate_action ("read-selected-article");
      else
        pane->_action_manager.activate_action ("save-attachments");
    }
    return false;
  }
}

void
HeaderPane :: on_row_activated (GtkTreeView          * treeview,
                                GtkTreePath          * path,
                                GtkTreeViewColumn    * col,
                                gpointer               pane_g)
{
  g_idle_add (on_row_activated_idle, pane_g);
}

/****
*****
*****  FILTERS
*****
****/

namespace
{
  const char * mode_strings [] =
  {
    _("Subject or Author"),
    _("Subject"),
    _("Author")
  };

  enum
  {
    SUBJECT_OR_AUTHOR,
    SUBJECT,
    AUTHOR
  };
}

void
HeaderPane :: rebuild_filter (const std::string& text, int mode)
{
  TextMatch::Description d;
  d.negate = false;
  d.case_sensitive = false;
  d.type = TextMatch::CONTAINS;
  d.text = text;

  FilterInfo &f (_filter);
  f.set_type_aggregate_and ();

  // entry field filter...
  FilterInfo entry_filter;
  if (!text.empty()) {
    if (mode == SUBJECT) {
      entry_filter.set_type_text ("Subject", d);
//std::cerr << LINE_ID << " AND subject contains [" << text << "]" << std::endl;
    } else if (mode == AUTHOR) {
//std::cerr << LINE_ID << " AND author contains [" << text << "]" << std::endl;
    } else { // mode == SUBJECT_OR_AUTHOR
      FilterInfo f1, f2;
      entry_filter.set_type_aggregate_or ();
      f1.set_type_text ("Subject", d);
      f2.set_type_text ("From", d);
//std::cerr << LINE_ID << " AND subject or author contains [" << text << "]" << std::endl;
      entry_filter._aggregates.push_back (f1);
      entry_filter._aggregates.push_back (f2);
    }
    f._aggregates.push_back (entry_filter);
  }

  if (_action_manager.is_action_active("match-only-unread-articles")) {
//std::cerr << LINE_ID << " AND is unread" << std::endl;
    FilterInfo tmp;
    tmp.set_type_is_unread ();
    f._aggregates.push_back (tmp);
  }
  if (_action_manager.is_action_active("match-only-cached-articles")) {
//std::cerr << LINE_ID << " AND is cached" << std::endl;
    FilterInfo tmp;
    tmp.set_type_cached ();
    f._aggregates.push_back (tmp);
  }
  if (_action_manager.is_action_active("match-only-binary-articles")) {
//std::cerr << LINE_ID << " AND has an attachment" << std::endl;
    FilterInfo tmp;
    tmp.set_type_binary ();
    f._aggregates.push_back (tmp);
  }
  if (_action_manager.is_action_active("match-only-my-articles")) {
//std::cerr << LINE_ID << " AND was posted by me" << std::endl;
    FilterInfo tmp;
    tmp.set_type_posted_by_me ();
    f._aggregates.push_back (tmp);
  }

  // try to fold the six ranges into as few FilterInfo items as possible..
  typedef std::pair<int,int> range_t;
  std::vector<range_t> ranges;
  ranges.reserve (6);
  if (_action_manager.is_action_active("match-ignored-articles")) ranges.push_back (range_t(INT_MIN, -9999));
  if (_action_manager.is_action_active("match-low-scoring-articles")) ranges.push_back (range_t(-9998, -1));
  if (_action_manager.is_action_active("match-normal-scoring-articles")) ranges.push_back (range_t(0, 0));
  if (_action_manager.is_action_active("match-medium-scoring-articles")) ranges.push_back (range_t(1, 4999));
  if (_action_manager.is_action_active("match-high-scoring-articles")) ranges.push_back (range_t(5000, 9998));
  if (_action_manager.is_action_active("match-watched-articles")) ranges.push_back (range_t(9999, INT_MAX));
  for (size_t i=0; i<ranges.size()-1; ) {
    if (ranges[i].second+1 != ranges[i+1].first)
      ++i;
    else {
      ranges[i].second = ranges[i+1].second;
      ranges.erase (ranges.begin()+i+1);
    }
  }

//for (size_t i=0; i<ranges.size(); ++i) std::cerr << LINE_ID << " range [" << ranges[i].first << "..." << ranges[i].second << "]" << std::endl;

  std::vector<FilterInfo> filters;
  for (size_t i=0; i<ranges.size(); ++i) {
    const range_t& range (ranges[i]);
    const bool low_bound (range.first == INT_MIN);
    const bool hi_bound (range.second == INT_MAX);
    if (low_bound && hi_bound) {
      // everything matches -- do nothing
    } else if (hi_bound) {
      FilterInfo tmp;
      tmp.set_type_score_ge (range.first);
//std::cerr << LINE_ID << " AND has a score >= " << range.first << std::endl;
      filters.push_back (tmp);
    } else if (low_bound) {
      FilterInfo tmp;
      tmp.set_type_score_le (range.second);
//std::cerr << LINE_ID << " AND has a score <= " << range.second << std::endl;
      filters.push_back (tmp);
    } else  { // not bound on either side; need an aggregate
      FilterInfo s, tmp;
      s.set_type_aggregate_and ();
      tmp.set_type_score_ge (range.first);
      s._aggregates.push_back (tmp);
      tmp.set_type_score_le (range.second);
      s._aggregates.push_back (tmp);
//std::cerr << LINE_ID << " AND has a in [" << range.first << "..." << range.second << ']' << std::endl;
      filters.push_back (s);
    }
  }
  if (filters.size()==1) // can fit in an `and' parent
    f._aggregates.push_back (filters[0]);
  else if (!filters.empty()) { // needs an `or' parent
    FilterInfo s;
    s.set_type_aggregate_or ();
    s._aggregates.swap (filters);
    f._aggregates.push_back (s);
  }
//std::cerr << LINE_ID << " number of filters: " << f._aggregates.size() << std::endl;
}

void
HeaderPane :: filter (const std::string& text, int mode)
{
  rebuild_filter (text, mode);

  if (_atree) {
    if (_filter._aggregates.empty())
      _atree->set_filter (0);
    else
      _atree->set_filter (&_filter);
  }
}

namespace
{
  // the text typed by the user.
  std::string search_text;

  guint entry_changed_tag (0u);
  guint activate_soon_tag (0u);

  // AUTHOR, SUBJECT, or SUBJECT_OR_AUTHOR
  int mode;

  void set_search_entry (GtkWidget * entry, const char * s)
  {
    g_signal_handler_block (entry, entry_changed_tag);
    gtk_entry_set_text (GTK_ENTRY(entry), s);
    g_signal_handler_unblock (entry, entry_changed_tag);
  }

  gboolean search_entry_focus_in_cb (GtkWidget     * w,
                                     GdkEventFocus * unused1,
                                     gpointer        unused2)
  {
    gtk_widget_modify_text (w, GTK_STATE_NORMAL, NULL); // resets
    set_search_entry (w, search_text.c_str());
    return false;
  }

  void refresh_search_entry (GtkWidget * w)
  {
    if (search_text.empty() && !GTK_WIDGET_HAS_FOCUS(w))
    {
      GdkColor c;
      c.pixel = 0;
      c.red = c.green = c.blue = 0xAAAA;
      gtk_widget_modify_text (w, GTK_STATE_NORMAL, &c);
      set_search_entry (w, mode_strings[mode]);
    }
  }

  gboolean search_entry_focus_out_cb (GtkWidget     * w,
                                      GdkEventFocus * unused1,
                                      gpointer        unused2)
  {
    refresh_search_entry (w);
    return false;
  }

  void search_activate (HeaderPane * h)
  {
    h->filter (search_text, mode);
  }

  void remove_activate_soon_tag ()
  {
    if (activate_soon_tag != 0)
    {
      g_source_remove (activate_soon_tag);
      activate_soon_tag = 0;
    }
  }

  void search_entry_activated (GtkEntry * entry, gpointer h_gpointer)
  {
    search_activate (static_cast<HeaderPane*>(h_gpointer));
    remove_activate_soon_tag ();
  }

  void reset_entry_filter_cb (GtkWidget * button, gpointer pane_gpointer)
  {
    GtkWidget * e = GTK_WIDGET (g_object_get_data (G_OBJECT(button), "entry"));
    set_search_entry (e, "");
    search_text.clear ();
    search_entry_activated (NULL, pane_gpointer);
  }

  gboolean activated_timeout_cb (gpointer h_gpointer)
  {
    search_activate (static_cast<HeaderPane*>(h_gpointer));
    remove_activate_soon_tag ();
    return false; // remove the source
  }

  // ensure there's exactly one activation timeout
  // and that it's set to go off one second from now.
  void bump_activate_soon_tag (HeaderPane * h)
  {
    remove_activate_soon_tag ();
    activate_soon_tag = g_timeout_add (1000, activated_timeout_cb, h);
  }

  // when the user changes the filter text,
  // update our state variable and bump the activate timeout.
  void search_entry_changed_by_user (GtkEditable * e, gpointer h_gpointer)
  {
    search_text = gtk_entry_get_text (GTK_ENTRY(e));
    bump_activate_soon_tag (static_cast<HeaderPane*>(h_gpointer));

    GtkWidget * b (GTK_WIDGET (g_object_get_data (G_OBJECT(e), "reset-button")));
    gtk_widget_set_sensitive (b, !search_text.empty());
  }

  // when the search mode is changed via the menu,
  // update our state variable and bump the activate timeout.
  void search_menu_toggled_cb (GtkCheckMenuItem  * menu_item,
                               gpointer            entry_g)
  {
    if (gtk_check_menu_item_get_active  (menu_item))
    {
      mode = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(menu_item), "MODE"));
      refresh_search_entry (GTK_WIDGET(entry_g));
      HeaderPane * h = (HeaderPane*) g_object_get_data (
                                             G_OBJECT(entry_g), "header-pane");
      bump_activate_soon_tag (h);
    }
  }

  // this pops up the `author, subject, subject-or-author' menu
  // when the search mode button is clicked.
  void search_mode_button_clicked_cb (GtkButton * button, gpointer menu_g)
  {
    gtk_menu_popup (GTK_MENU(menu_g), 0, 0,
                    0, 0,
                    0, gtk_get_current_event_time());
  }
}

void
HeaderPane :: refilter ()
{
  search_activate (this);
}

HeaderPane :: ~HeaderPane ()
{
  _prefs.remove_listener (this);
  _data.remove_listener (this);

  set_group (Quark());
}

HeaderPane :: HeaderPane (ActionManager       & action_manager,
                          Data                & data,
                          const ArticleCache  & cache,
                          Prefs               & prefs):
  _action_manager (action_manager),
  _data (data),
  _prefs (prefs),
  _atree (0),
  _root (0),
  _entry (0),
  _tree_view (0),
  _toolbar_bin (gtk_event_box_new ()),
  _tree_store (0),
  _cache (cache)
{
  // init the icons
  for (guint i=0; i<ICON_QTY; ++i)
    _icons[i].pixbuf = gdk_pixbuf_new_from_inline (-1, _icons[i].pixbuf_txt, FALSE, 0);

  _entry = gtk_entry_new ();
  _action_manager.disable_accelerators_when_focused (_entry);
  g_object_set_data (G_OBJECT(_entry), "header-pane", this);
  g_signal_connect (_entry, "focus-in-event", G_CALLBACK(search_entry_focus_in_cb), NULL);
  g_signal_connect (_entry, "focus-out-event", G_CALLBACK(search_entry_focus_out_cb), NULL);
  g_signal_connect (_entry, "activate", G_CALLBACK(search_entry_activated), this);
  entry_changed_tag = g_signal_connect (_entry, "changed", G_CALLBACK(search_entry_changed_by_user), this);
  refresh_search_entry (_entry);
  GtkWidget * menu = gtk_menu_new ();
  mode = 0;
  GSList * l = 0;
  for (int i=0, qty=G_N_ELEMENTS(mode_strings); i<qty; ++i) {
    GtkWidget * w = gtk_radio_menu_item_new_with_label (l, mode_strings[i]);
    l = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM(w));
    g_object_set_data (G_OBJECT(w), "MODE", GINT_TO_POINTER(i));
    g_signal_connect (w, "toggled", G_CALLBACK(search_menu_toggled_cb),_entry);
    gtk_menu_shell_append (GTK_MENU_SHELL(menu), w);
    gtk_widget_show (w);
  }

  GtkWidget * search_mode_button = gtk_button_new ();
  g_signal_connect (search_mode_button, "clicked", G_CALLBACK(search_mode_button_clicked_cb), menu);
  GtkWidget * image = gtk_image_new_from_stock ("ICON_SEARCH_PULLDOWN", GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER(search_mode_button), image);
  gtk_button_set_relief (GTK_BUTTON(search_mode_button), GTK_RELIEF_NONE);

  GtkTooltips * tips = gtk_tooltips_new ();
  GtkWidget * w = gtk_button_new ();
  g_object_set_data (G_OBJECT(w), "entry", _entry);
  g_object_set_data (G_OBJECT(_entry), "reset-button", w);
  gtk_button_set_relief (GTK_BUTTON(w), GTK_RELIEF_NONE);
  g_signal_connect (w, "clicked", G_CALLBACK(reset_entry_filter_cb), this);
  image = gtk_image_new_from_stock (GTK_STOCK_CANCEL, GTK_ICON_SIZE_BUTTON);
  gtk_container_add (GTK_CONTAINER(w), image);
  gtk_tooltips_set_tip (GTK_TOOLTIPS(tips), w, _("Clear the Filter"), NULL);
  gtk_widget_set_sensitive (w, false);

  GtkWidget * toolbar = _action_manager.get_action_widget ("/header-pane-toolbar");
  assert (toolbar->parent == 0);
  gtk_toolbar_set_icon_size (GTK_TOOLBAR(toolbar), GTK_ICON_SIZE_BUTTON);
  gtk_toolbar_set_style (GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
  GtkToolItem * item = gtk_separator_tool_item_new ();
  gtk_widget_show (GTK_WIDGET(item));
  gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

  item = gtk_tool_item_new ();
  gtk_container_add (GTK_CONTAINER(item), search_mode_button);
  gtk_widget_show_all (GTK_WIDGET(item));
  gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

  item = gtk_tool_item_new ();
  gtk_container_add (GTK_CONTAINER(item), _entry);
  gtk_widget_show_all (GTK_WIDGET(item));
  gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

  item = gtk_tool_item_new ();
  gtk_container_add (GTK_CONTAINER(item), w);
  gtk_widget_show_all (GTK_WIDGET(item));
  gtk_toolbar_insert (GTK_TOOLBAR(toolbar), item, -1);

  gtk_widget_show (GTK_WIDGET(toolbar));
  gtk_widget_show (_toolbar_bin);
  gtk_container_add (GTK_CONTAINER(_toolbar_bin), GTK_WIDGET(toolbar));


  // build the view...
  _tree_view = gtk_tree_view_new ();
  gtk_tree_view_set_enable_search (GTK_TREE_VIEW(_tree_view), false);
  GtkTreeSelection * sel = gtk_tree_view_get_selection (GTK_TREE_VIEW(_tree_view));
  gtk_tree_selection_set_mode (sel, GTK_SELECTION_MULTIPLE);
#if GTK_CHECK_VERSION(2,6,0)
  gtk_tree_view_set_fixed_height_mode (GTK_TREE_VIEW(_tree_view), true);
#endif

  g_signal_connect (_tree_view, "button-release-event", G_CALLBACK(on_button_pressed), this);
  g_signal_connect (_tree_view, "button-press-event", G_CALLBACK(on_button_pressed), this);
  g_signal_connect (_tree_view, "popup-menu", G_CALLBACK(on_popup_menu), this);
  g_signal_connect (_tree_view, "row-activated", G_CALLBACK(on_row_activated), this);
  GtkWidget * scroll = gtk_scrolled_window_new (0, 0);
  gtk_container_add (GTK_CONTAINER(scroll), _tree_view);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW(scroll), GTK_SHADOW_IN);


  GtkCellRenderer * pixbuf_renderer = GTK_CELL_RENDERER (
    g_object_new (GTK_TYPE_CELL_RENDERER_PIXBUF,
    "ypad", 0,
    "xpad", 1, NULL));
  GtkCellRenderer * text_renderer = GTK_CELL_RENDERER (
    g_object_new (GTK_TYPE_CELL_RENDERER_TEXT,
    "ypad", 0,
    "xpad", 1, NULL));

  // state
  GtkTreeViewColumn * col = gtk_tree_view_column_new ();
  gtk_tree_view_column_set_sizing (col, GTK_TREE_VIEW_COLUMN_FIXED);
  gtk_tree_view_column_set_fixed_width (col, 24);
  gtk_tree_view_column_set_resizable (col, false);
  gtk_tree_view_column_pack_start (col, pixbuf_renderer, false);
  gtk_tree_view_column_set_cell_data_func (col, pixbuf_renderer, render_state, 0, 0);
  gtk_tree_view_column_set_sort_column_id (col, COL_STATE);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_tree_view), col);

  // action
  col = gtk_tree_view_column_new ();
  gtk_tree_view_column_set_sizing (col ,GTK_TREE_VIEW_COLUMN_FIXED);
  gtk_tree_view_column_set_fixed_width (col, 24);
  gtk_tree_view_column_set_resizable (col, false);
  gtk_tree_view_column_pack_start (col, pixbuf_renderer, false);
  gtk_tree_view_column_set_cell_data_func (col, pixbuf_renderer, render_action, this, 0);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_tree_view), col);

  // subject
  col = gtk_tree_view_column_new_with_attributes (_("Subject"), text_renderer, NULL);
  gtk_tree_view_column_set_sizing (col ,GTK_TREE_VIEW_COLUMN_FIXED);
  gtk_tree_view_column_set_fixed_width (col, 400);
  gtk_tree_view_column_set_resizable (col, true);
  gtk_tree_view_column_set_sort_column_id (col, COL_SUBJECT);
  gtk_tree_view_column_set_cell_data_func (col, text_renderer, render_subject, &_data, 0);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_tree_view), col);
  gtk_tree_view_set_expander_column (GTK_TREE_VIEW(_tree_view), col);

  // score
  GtkCellRenderer * score_renderer = GTK_CELL_RENDERER (g_object_new (
    GTK_TYPE_CELL_RENDERER_TEXT,
    "ypad", 0,
    "xpad", 1, NULL));
  col = gtk_tree_view_column_new_with_attributes (_("Score"), score_renderer, NULL);
  gtk_tree_view_column_set_sizing (col, GTK_TREE_VIEW_COLUMN_FIXED);
  gtk_tree_view_column_set_fixed_width (col, 50);
  gtk_tree_view_column_set_resizable (col, true);
  gtk_tree_view_column_set_sort_column_id (col, COL_SCORE);
  gtk_tree_view_column_set_cell_data_func (col, score_renderer, render_score, &_prefs, 0);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_tree_view), col);

  // author
  col = gtk_tree_view_column_new_with_attributes (_("Author"), text_renderer, NULL);
  gtk_tree_view_column_set_sizing (col, GTK_TREE_VIEW_COLUMN_FIXED);
  gtk_tree_view_column_set_fixed_width (col, 133);
  gtk_tree_view_column_set_resizable (col, true);
  gtk_tree_view_column_set_sort_column_id (col, COL_AUTHOR);
  gtk_tree_view_column_set_cell_data_func (col, text_renderer, render_author, 0, 0);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_tree_view), col);

  // lines
  col = gtk_tree_view_column_new_with_attributes (_("Lines"), text_renderer, NULL);
  gtk_tree_view_column_set_sizing (col, GTK_TREE_VIEW_COLUMN_FIXED);
  gtk_tree_view_column_set_fixed_width (col, 60);
  gtk_tree_view_column_set_resizable (col, true);
  gtk_tree_view_column_set_sort_column_id (col, COL_LINES);
  gtk_tree_view_column_set_cell_data_func (col, text_renderer, render_lines, 0, 0);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_tree_view), col);

  // date
  GtkCellRenderer * date_renderer = GTK_CELL_RENDERER (g_object_new (
    GTK_TYPE_CELL_RENDERER_TEXT, 
    "ypad", 0,
    "xpad", 1, NULL));
  col = gtk_tree_view_column_new_with_attributes (_("Date"), date_renderer, "text", COL_DATE_STR, NULL);
  gtk_tree_view_column_set_sizing (col, GTK_TREE_VIEW_COLUMN_FIXED);
  gtk_tree_view_column_set_fixed_width (col, 120);
  gtk_tree_view_column_set_resizable (col, false);
  gtk_tree_view_column_set_sort_column_id (col, COL_DATE);
  gtk_tree_view_append_column (GTK_TREE_VIEW(_tree_view), col);

  // join rows
  _root = gtk_vbox_new (false, PAD_SMALL);
  gtk_box_pack_start (GTK_BOX(_root), _toolbar_bin, false, false, 0);
  gtk_box_pack_start (GTK_BOX(_root), scroll, true, true, 0);

  search_activate (this); // calls rebuild_filter

  _data.add_listener (this);
  _prefs.add_listener (this);

  refresh_font ();
  refresh_color ();
}

void
HeaderPane :: select_all ()
{
  GtkTreeSelection * sel (gtk_tree_view_get_selection (GTK_TREE_VIEW (_tree_view)));
  gtk_tree_selection_select_all (sel);
}

void
HeaderPane :: unselect_all ()
{
  GtkTreeSelection * sel (gtk_tree_view_get_selection (GTK_TREE_VIEW (_tree_view)));
  gtk_tree_selection_unselect_all (sel);
}

namespace
{
  void select_subpaths (GtkTreeView * view, GtkTreeModel * model, GtkTreeSelection * sel, GtkTreePath * begin_path)
  {
    gtk_tree_view_expand_row (view, begin_path, true);
    GtkTreeIter begin;
    gtk_tree_model_get_iter (model, &begin, begin_path);
    GtkTreeIter end = begin;
    int n;
    while ((n = gtk_tree_model_iter_n_children (model, &end))) {
      GtkTreeIter tmp = end;
      gtk_tree_model_iter_nth_child (model, &end, &tmp, n-1);
    }
    GtkTreePath * end_path = gtk_tree_model_get_path (model,  &end);
    gtk_tree_selection_select_range (sel, begin_path, end_path);
    gtk_tree_path_free (end_path);
  }

  void select_threads_helper (GtkTreeView * view, bool subthreads_only)
  {
    GtkTreeModel * model;
    GtkTreeSelection * sel (gtk_tree_view_get_selection (view));
    GList * list (gtk_tree_selection_get_selected_rows (sel, &model));

    for (GList *l(list); l; l=l->next) {
      GtkTreePath * path (static_cast<GtkTreePath*>(l->data));
      if (!subthreads_only)
        while (gtk_tree_path_get_depth (path) > 1) // up to root
        gtk_tree_path_up (path);
      select_subpaths (view, model, sel, path);
    }

    g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (list);
  }
}

void
HeaderPane :: expand_first_selected_article ()
{
  // get a list of paths
  GtkTreeModel * model (0);
  GtkTreeSelection * selection (gtk_tree_view_get_selection (GTK_TREE_VIEW(_tree_view)));
  GList * list (gtk_tree_selection_get_selected_rows (selection, &model));
  GtkTreePath * path = list ? (static_cast<GtkTreePath*>(list->data)) : 0;

  gtk_tree_view_expand_row (GTK_TREE_VIEW(_tree_view), path, true);

  // cleanup
  g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
  g_list_free (list);
}

void
HeaderPane :: select_threads ()
{
  select_threads_helper (GTK_TREE_VIEW(_tree_view), false);
}

void
HeaderPane :: select_subthreads ()
{
  select_threads_helper (GTK_TREE_VIEW(_tree_view), true);
}

/***
****
****  NAVIGATION
****
***/

namespace
{
  /**
  ***  Navigation Functors
  **/

  struct TreeIterFunctor
  {
    virtual ~TreeIterFunctor () {}
    virtual bool operator() (GtkTreeModel*, GtkTreeIter*) const = 0;
    virtual bool front (GtkTreeModel*, GtkTreeIter*) const = 0;
  };

  struct TreeIteratorNext: public TreeIterFunctor
  {
    virtual ~TreeIteratorNext () {}

    virtual bool operator ()(GtkTreeModel * model, GtkTreeIter * iter) const
    {
      g_assert (model != 0);
      g_assert (iter != 0);

      // childen
      GtkTreeIter tmp;
      if (gtk_tree_model_iter_nth_child (model, &tmp, iter, 0)) {
        *iter = tmp;
        return true;
      }

      // siblings
      GtkTreeIter up = *iter;
      do {
        if (gtk_tree_model_iter_next (model, &up)) {
          *iter = up;
          return true;
        }
        tmp = up;
      } while (gtk_tree_model_iter_parent (model, &up, &tmp));

      return false;
    }

    virtual bool front (GtkTreeModel* model, GtkTreeIter* setme) const {
      return gtk_tree_model_get_iter_first (model, setme);
    }
  };

  struct TreeIteratorPrev: public TreeIterFunctor
  {
    private:

      // only works for iterators of a tree store.
      // treestore's iter->user_data is a GNode pointer...
      bool
      tree_store_iter_equal (const GtkTreeIter * a, const GtkTreeIter * b) const
      {
        if (!a && !b) return true;
        if (!a || !b) return false;
        return a->user_data == b->user_data;
      }

      void
      tree_model_iter_last_descendant (GtkTreeModel * model,
                                       GtkTreeIter * setme,
                                       GtkTreeIter * parent) const
      {
        GtkTreeIter march (*parent);
        for (;;) {
          const int n_children (gtk_tree_model_iter_n_children (model, &march));
          if (!n_children) {
            *setme = march;
            return;
          }
          GtkTreeIter tmp;
          gtk_tree_model_iter_nth_child (model, &tmp, &march, n_children-1);
          march = tmp;
        }
      }

    public:

      virtual ~TreeIteratorPrev () {}

      bool
      operator()(GtkTreeModel * model, GtkTreeIter * iter) const
      {
        // get parent
        GtkTreeIter parent;
        const bool has_parent = gtk_tree_model_iter_parent (model, &parent, iter);

        // get sibling
        GtkTreeIter sibling;
        gtk_tree_model_iter_children (model, &sibling, (has_parent ? &parent : 0));
        if (tree_store_iter_equal (&sibling, iter)) { // iter is the first child
          *iter = parent;
          return has_parent;
        }

        // walk along siblings until we find the one prior to iter.
        // gtk_tree_model_iter_prev() would have been nice here.
        for (;;) {
          GtkTreeIter tmp = sibling;
          g_assert (gtk_tree_model_iter_next (model, &tmp));
          if (tree_store_iter_equal (&tmp, iter))
            break;
          sibling = tmp;
        }

        // find the absolutely last child of the older sibling
        tree_model_iter_last_descendant (model, iter, &sibling);
        return true;
      }

      // aka gtk_tree_model_get_iter_last (model, setme);
      virtual bool front (GtkTreeModel* model, GtkTreeIter* setme) const {
        const int n (gtk_tree_model_iter_n_children (model, NULL));
        GtkTreeIter tmp;
        if (!gtk_tree_model_iter_nth_child (model, &tmp, NULL, n-1))
          return false;
        tree_model_iter_last_descendant (model, setme, &tmp);
        return true;
      }
  };

  /**
  ***  Article Test Functors
  **/

  struct ArticleExists: public ArticleFunctor {
    virtual ~ArticleExists() {}
    ArticleExists () {}
    virtual bool operator()(const Article& a) const { return true; }
  };

  struct ArticleIsParentOf: public ArticleFunctor {
    virtual ~ArticleIsParentOf () {}
    ArticleIsParentOf (const Data::ArticleTree& tree, const Article* a) {
      const Article * parent = a ? tree.get_parent(a->message_id) : 0;
      _mid = parent ? parent->message_id : "";
    }
    Quark _mid;
    virtual bool operator()(const Article& a) const { return _mid==a.message_id; }
  };

  struct ArticleIsUnread: public ArticleFunctor {
    virtual ~ArticleIsUnread () {}
    ArticleIsUnread (const Data& data): _data(data) {}
    const Data& _data;
    virtual bool operator()(const Article& a) const { return !_data.is_read(&a); }
  };

  struct ArticleIsNotInThread: public ArticleFunctor {
    virtual ~ArticleIsNotInThread () {}
    ArticleIsNotInThread (const Data::ArticleTree& tree, const Article* a):
      _tree(tree),
      _root(a ? get_root_mid(a) : "") {}
    const Data::ArticleTree& _tree;
    const Quark _root;
    Quark get_root_mid (const Article * a) const {
      for (;;) {
        const Quark mid (a->message_id);
        const Article * parent = _tree.get_parent (mid);
        if (!parent)
          return mid;
        a = parent;
      }
    }
    virtual bool operator()(const Article& a) const {
      return _root != get_root_mid(&a);
    }
  };

  struct ArticleIsUnreadAndNotInThread: public ArticleFunctor {
    virtual ~ArticleIsUnreadAndNotInThread () {}
    ArticleIsUnreadAndNotInThread (const Data& data, const Data::ArticleTree& tree, const Article* a): _aiu (data), _ainit(tree, a) {}
    const ArticleIsUnread _aiu;
    const ArticleIsNotInThread _ainit;
    virtual bool operator()(const Article& a) const {
      return _aiu(a) && _ainit(a);
    }
  };

  /**
  ***  Action Functors
  **/

  struct ActionFunctor {
    virtual ~ActionFunctor () {}
    virtual void operator() (GtkTreeModel* model, GtkTreeIter* iter, const Article& a) const = 0;
  };

  struct SelectFunctor: public ActionFunctor {
    virtual ~SelectFunctor () {}
    SelectFunctor (GtkTreeView * view): _view(view) {}
    GtkTreeView * _view;
    virtual void operator() (GtkTreeModel* model, GtkTreeIter* iter, const Article& a) const {
      GtkTreeSelection * sel (gtk_tree_view_get_selection (_view));
      gtk_tree_selection_unselect_all (sel);
      GtkTreePath * path = gtk_tree_model_get_path (model, iter);
      gtk_tree_view_expand_row (_view, path, true);
      gtk_tree_view_expand_to_path (_view, path);
      gtk_tree_view_set_cursor (_view, path, NULL, FALSE);
      gtk_tree_view_scroll_to_cell (_view, path, NULL, true, 0.5f, 0.0f);
      gtk_tree_path_free (path);
    }
  };

  struct ReadFunctor: public SelectFunctor {
    virtual ~ReadFunctor() {}
    ReadFunctor (GtkTreeView * view, ActionManager& am): SelectFunctor(view), _am(am) {}
    ActionManager& _am;
    virtual void operator() (GtkTreeModel* model, GtkTreeIter* iter, const Article& a) const {
      SelectFunctor::operator() (model, iter, a);
      _am.activate_action ("read-selected-article");
    }
  };

  /**
  *** 
  **/

  void
  find_next_iterator_from (GtkTreeModel               * model,
                           GtkTreeIter                * start_pos,
                           const TreeIterFunctor      & iterate_func,
                           const ArticleFunctor       & test_func,
                           const ActionFunctor        & success_func,
                           bool                         test_the_start_pos)
  {
    g_assert (start_pos!=0);
    GtkTreeIter march = *start_pos;
    bool success (false);
    const Article *article (0);
    for (;;)
    {
      if (test_the_start_pos)
        test_the_start_pos = false;
      else if (!iterate_func (model, &march))
        break;
      article = get_rec (model, &march).article;
      g_assert (article != 0);
      if ((success = test_func (*article)))
        break;
    }
    if (success)
      success_func (model, &march, *article);
  }

  bool get_first_selection (GtkTreeSelection * sel, GtkTreeIter * setme)
  {
    GtkTreeModel * model (0);
    GList * list (gtk_tree_selection_get_selected_rows (sel, &model));
    const bool found (list && gtk_tree_model_get_iter (model, setme, (GtkTreePath*)(list->data)));
    g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
    g_list_free (list);
    return found;
  }

  void
  next_iterator (GtkTreeView                 * view,
                 GtkTreeModel                * model,
                 const TreeIterFunctor       & iterate_func,
                 const ArticleFunctor        & test_func,
                 const ActionFunctor         & success_func)
  {
    GtkTreeSelection * sel = gtk_tree_view_get_selection (view);
    GtkTreeIter iter;
    if (get_first_selection (sel, &iter))
      find_next_iterator_from (model, &iter, iterate_func, test_func, success_func, false);
    else if (iterate_func.front (model, &iter))
      find_next_iterator_from (model, &iter, iterate_func, test_func, success_func, true);
  }
}

void
HeaderPane :: read_next_if (const ArticleFunctor& test)
{
  GtkTreeView * v (GTK_TREE_VIEW(_tree_view));
  GtkTreeModel * m (GTK_TREE_MODEL(_tree_store));
  next_iterator (v, m, TreeIteratorNext(), test, ReadFunctor(v,_action_manager));
}

void
HeaderPane :: read_prev_if (const ArticleFunctor & test)
{
  GtkTreeView * v (GTK_TREE_VIEW(_tree_view));
  GtkTreeModel * m (GTK_TREE_MODEL(_tree_store));
  next_iterator (v, m, TreeIteratorPrev(), test, ReadFunctor(v,_action_manager));
}

void
HeaderPane :: read_next_unread_article ()
{
  read_next_if (ArticleIsUnread(_data));
}

void
HeaderPane :: read_previous_article ()
{
  read_prev_if (ArticleExists());
}

void
HeaderPane :: read_next_article ()
{
  read_next_if (ArticleExists());
}

void
HeaderPane :: read_next_thread ()
{
  if (_atree)
    read_next_if (ArticleIsNotInThread(*_atree, get_first_selected_article()));
}

void
HeaderPane :: read_next_unread_thread ()
{
  if (_atree)
    read_next_if (ArticleIsUnreadAndNotInThread(_data, *_atree, get_first_selected_article()));
}

void
HeaderPane :: read_previous_thread ()
{
  if (_atree)
    read_prev_if (ArticleIsNotInThread(*_atree, get_first_selected_article()));
}

void
HeaderPane :: read_parent_article ()
{
  if (_atree)
    read_prev_if (ArticleIsParentOf(*_atree, get_first_selected_article()));
}

/***
****
***/

void
HeaderPane :: refresh_font ()
{
  if (!_prefs.get_flag ("header-pane-font-enabled", false))
    gtk_widget_modify_font (_tree_view, 0);
  else {
    const std::string str (_prefs.get_string ("header-pane-font", "Sans 10"));
    PangoFontDescription * pfd (pango_font_description_from_string (str.c_str()));
    gtk_widget_modify_font (_tree_view, pfd);
    pango_font_description_free (pfd);
  }
}

void
HeaderPane :: on_prefs_flag_changed (const StringView& key, bool value)
{
  if (key == "header-pane-font-enabled")
    refresh_font ();
}

void
HeaderPane :: on_prefs_string_changed (const StringView& key, const StringView& value)
{
  if (key == "header-pane-font")
    refresh_font ();
}

void
HeaderPane :: refresh_color ()
{
}

void
HeaderPane :: on_prefs_color_changed (const StringView& key, const GdkColor& color)
{
  if (key == "header-pane-background")
    refresh_color ();
}

