/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdarg>
#include <gobject/gvaluecollector.h>
#include <pan/general/debug.h>
#include "pan-tree.h"

#define IS_SORTED(tree) (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID)

namespace
{
  void set_iter (GtkTreeIter   * iter,
                 int             stamp,
                 PanTreeStore::Node * node)
  {
    iter->stamp = stamp;
    iter->user_data = node;
    iter->user_data2 = 0;
    iter->user_data3 = 0;
  }

  void invalidate_iter (GtkTreeIter * iter)
  {
    set_iter (iter, 0, 0);
  }

  GtkTreeModelFlags get_flags (GtkTreeModel *tree_model)
  {
    return GTK_TREE_MODEL_ITERS_PERSIST;
  }

  gint get_n_columns (GtkTreeModel *tree)
  {
    return PAN_TREE_STORE(tree)->n_columns;
  }

  GType get_column_type (GtkTreeModel * tree,
                            gint           n)
  {
    return (*((PanTreeStore*)(tree))->column_types)[n];
  }

  gboolean get_iter (GtkTreeModel *tree_model,
                     GtkTreeIter  *iter,
                     GtkTreePath  *path)
  {
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (path, false);

    // find the article that corresponds to this path.
    PanTreeStore::Node * node = tree->root;
    const int depth = gtk_tree_path_get_depth (path);
    const int * indices = gtk_tree_path_get_indices (path);
    for (int i=0; i<depth; ++i) {
      const int index = *indices++;
      if (index<0 || index>=node->n_children)
        return false;
      node = node->children[index];
    }

    set_iter (iter, tree->stamp, node);
    return true;
  }

  GtkTreePath* get_path (GtkTreeModel *tree_model,
                         GtkTreeIter  *iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (iter->stamp == tree->stamp, false);

    GtkTreePath * path = gtk_tree_path_new ();
    PanTreeStore::Node * node = static_cast<PanTreeStore::Node*>(iter->user_data);
    while (node && node!=tree->root) {
      gtk_tree_path_prepend_index (path, node->child_index);
      node = node->parent;
    }
//std::cerr << LINE_ID << " got tree path " << gtk_tree_path_to_string(path) << std::endl;
    return path;
  }

  void get_value (GtkTreeModel * tree_model,
                  GtkTreeIter  * iter,
                  gint           column,
                  GValue       * dest_value)
  {
    g_assert (iter);
    g_assert (dest_value);
    PanTreeStore * tree = PAN_TREE_STORE(tree_model);
    g_assert (tree);
    g_assert (iter->stamp == tree->stamp);
    g_assert (0<=column && column<tree->n_columns);

    PanTreeStore::Node * node = static_cast<PanTreeStore::Node*>(iter->user_data);
    g_value_init (dest_value, (*tree->column_types)[column]);
    g_value_copy (&node->columns[column], dest_value);
  }

  gboolean iter_next (GtkTreeModel * model,
                      GtkTreeIter  * iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (iter->stamp == tree->stamp, false);

    PanTreeStore::Node * node = static_cast<PanTreeStore::Node*>(iter->user_data);
    const int new_index = node->child_index + 1;
    if (new_index >= node->parent->n_children) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->parent->children[new_index]);
    return true;
  }

  gboolean iter_children (GtkTreeModel * model,
                          GtkTreeIter  * iter,
                          GtkTreeIter  * parent)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (parent->stamp == tree->stamp, false);

    PanTreeStore::Node * node = parent ? static_cast<PanTreeStore::Node*>(parent->user_data) : tree->root;
    if (node->children.empty()) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->children[0]);
    return true;
  }

  gint iter_n_children (GtkTreeModel * model,
                        GtkTreeIter  * iter)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, 0);
    g_return_val_if_fail (!iter || iter->stamp == tree->stamp, 0);
    if (iter) g_assert (iter->user_data != tree->root);

    PanTreeStore::Node * node = iter ? static_cast<PanTreeStore::Node*>(iter->user_data) : tree->root;
    return node->n_children;
  }

  gboolean iter_has_child (GtkTreeModel *model,
                           GtkTreeIter  *iter)
  {
    return iter_n_children (model, iter) != 0;
  }

  gboolean iter_nth_child (GtkTreeModel * model,
                           GtkTreeIter  * iter,
                           GtkTreeIter  * parent,
                           gint           n)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (!parent || parent->stamp == tree->stamp, false);

    PanTreeStore::Node * node = parent ? static_cast<PanTreeStore::Node*>(parent->user_data) : tree->root;
    if (n<0 || n>=node->n_children) {
      invalidate_iter (iter);
      return false;
    }

    set_iter (iter, tree->stamp, node->children[n]);
    //std::cerr << LINE_ID << " getting child # " << n << " of " << iter << " returning true" << std::endl;
    return true;
  }

  gboolean iter_parent (GtkTreeModel * model,
                        GtkTreeIter  * iter,
                        GtkTreeIter  * child)
  {
    PanTreeStore * tree = PAN_TREE_STORE(model);
    g_return_val_if_fail (tree, false);
    g_return_val_if_fail (child, false);
    g_return_val_if_fail (iter, false);
    g_return_val_if_fail (child->stamp == tree->stamp, false);

    PanTreeStore::Node * node = static_cast<PanTreeStore::Node*>(child->user_data);
    const bool has_parent = node->parent != tree->root;
    if (has_parent)
      set_iter (iter, tree->stamp, node->parent);
    else
      invalidate_iter (iter);
    return has_parent;
  }
}

namespace
{
  gboolean get_sort_column_id (GtkTreeSortable  * sortable,
                               gint             * sort_column_id,
                               GtkSortType      * order)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_val_if_fail (tree, false);

    if (sort_column_id)
       *sort_column_id = tree->sort_column_id;
    if (order)
       *order = tree->order;

    return (tree->sort_column_id != GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID)
        && (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID);
  }

  gboolean has_sort_func (GtkTreeSortable *sortable, int col)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_val_if_fail (tree, false);
    std::map<int,PanTreeStore::SortInfo>::const_iterator it = tree->sort_info->find (col);
    return (it!=tree->sort_info->end()) && (it->second.sort_func!=0);
  }

  struct SortTuple
  {
    int offset;
    PanTreeStore::Node * node;
    GtkTreeIter iter;
  };

  struct SortHelperData
  {
    GtkTreeModel * model;
    PanTreeStore::SortInfo& sort_info;
    GtkSortType order;
    SortHelperData (GtkTreeModel * in_model, PanTreeStore::SortInfo& in_sort_info, GtkSortType in_order):
      model     (in_model),
      sort_info (in_sort_info),
      order     (in_order) {}
  };

  int tuple_compare_func (gconstpointer a_gpointer,
                          gconstpointer b_gpointer,
                          gpointer user_data)
  {
    const SortTuple * a = (const SortTuple*) a_gpointer;
    const SortTuple * b = (const SortTuple*) b_gpointer;
    const SortHelperData * help = (const SortHelperData*) user_data;
    const int val = (help->sort_info.sort_func) (help->model,
                                                 const_cast<GtkTreeIter*>(&a->iter),
                                                 const_cast<GtkTreeIter*>(&b->iter),
                                                 help->sort_info.user_data);
    return help->order==GTK_SORT_ASCENDING ? val : -val;
  }

  void pan_tree_store_sort_helper (PanTreeStore            * tree,
                                   PanTreeStore::SortInfo  & sort_info,
                                   PanTreeStore::Node      * parent,
                                   bool                 recurse)
  {
    if (parent->n_children < 2) // no need to sort [0...1] children
      return;

    const int n = parent->n_children;

    // build a temporary array to sort
    std::vector<SortTuple> sort_array (n);
    for (int i=0; i<n; ++i) {
      SortTuple& tuple = sort_array[i];
      tuple.offset = i;
      tuple.node = parent->children[i];
      set_iter (&tuple.iter, tree->stamp, tuple.node);
    }

    // sort 'em...
    SortHelperData help (GTK_TREE_MODEL(tree), sort_info, tree->order);
    g_qsort_with_data (&sort_array.front(), sort_array.size(), sizeof(SortTuple), tuple_compare_func, &help);

    // update the child indices...
    for (int i=0; i<n; ++i) {
      PanTreeStore::Node * child = sort_array[i].node;
      child->child_index = i;
      parent->children[i] = child;
      g_assert (child->parent == parent);
    }

    // let the world know we've changed
    int * new_order = g_new (gint, n);
    bool differ = false;
    for (int i=0; i<n; ++i) {
      differ |= (i != sort_array[i].offset);
      new_order[i] = sort_array[i].offset;
    }

    // tell the world...
    GtkTreeIter tmp_iter, *parent_iter;
    GtkTreePath * path;
    if (parent == tree->root) {
      parent_iter = 0;
      path = gtk_tree_path_new ();
    } else {
      set_iter (&tmp_iter, tree->stamp, parent);
      parent_iter = &tmp_iter;
      path = gtk_tree_model_get_path (help.model, parent_iter);
    }
    gtk_tree_model_rows_reordered (help.model, path, parent_iter, new_order);
    gtk_tree_path_free (path);
    g_free (new_order);

#if 0
    // recurse?
    for (int i=0; recurse && i<n; ++i) {
      PanTreeStore::Node * child = parent->children[i];
      pan_tree_store_sort_helper (tree, sort_info, child, recurse);
    }
#endif
  }

  void pan_tree_store_sort (PanTreeStore * tree)
  {
    g_assert (IS_PAN_TREE_STORE(tree));

    if (tree->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID)
    {
      g_assert (has_sort_func (GTK_TREE_SORTABLE(tree), tree->sort_column_id));
      PanTreeStore::SortInfo& sort_info = (*tree->sort_info)[tree->sort_column_id];
      pan_tree_store_sort_helper (tree, sort_info, tree->root, true);
    }
  }

  void pan_tree_store_sort_iter_changed (PanTreeStore      * tree,
                                         GtkTreeIter  * iter,
                                         gint           column)
  {
    std::cerr << LINE_ID << " fixme" << std::endl;
  }


  void set_sort_column_id (GtkTreeSortable * sortable,
                           gint              sort_column_id,
                           GtkSortType       order)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_if_fail (tree);

    // if no change, there's nothing to do...
    if (tree->sort_column_id==sort_column_id && tree->order==order)
      return;

    // sanity checks...
    if (sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID) {
      g_return_if_fail (tree->sort_info->count(sort_column_id) != 0);
      g_return_if_fail (tree->sort_info->find(sort_column_id)->second.sort_func != 0);
    }

    tree->sort_column_id = sort_column_id;
    tree->order = order;
    gtk_tree_sortable_sort_column_changed (sortable);
    pan_tree_store_sort (tree);
  }

  gboolean has_default_sort_func (GtkTreeSortable *sortable)
  {
    return has_sort_func (sortable, GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID);
  }

  void set_sort_func (GtkTreeSortable        *sortable,
                      gint                    col,
                      GtkTreeIterCompareFunc  func,
                      gpointer                data,
                      GtkDestroyNotify        destroy)
  {
    PanTreeStore * tree (PAN_TREE_STORE (sortable));
    g_return_if_fail (tree);

    (*tree->sort_info)[col].assign (func, data, destroy);

    if (tree->sort_column_id == col)
      pan_tree_store_sort (tree);
  }

  void set_default_sort_func (GtkTreeSortable        *sortable,
                              GtkTreeIterCompareFunc  func,
                              gpointer                data,
                              GtkDestroyNotify        destroy)
  {
    set_sort_func (sortable, GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID, func, data, destroy);
  }
}

namespace
{
  GObjectClass *parent_class (0);

  typedef std::vector<GValue> values_t;

  void pan_tree_node_unset_values (PanTreeStore * tree, PanTreeStore::Node * node)
  {
    if (node->columns)
    {
      GValue * it = node->columns;
      for (int i=0, n=tree->n_columns; i<n; ++i)
        g_value_unset (it++);
      g_free (node->columns);
      node->columns = 0;
    }
  }

  void
  pan_tree_finalize (GObject *object)
  {
    PanTreeStore * tree = (PanTreeStore*)object;
    for (std::deque<PanTreeStore::Node>::iterator it(tree->nodes->begin()), end(tree->nodes->end()); it!=end; ++it)
      pan_tree_node_unset_values (tree, &*it);
    
    delete tree->sort_info;
    delete tree->column_types;
    delete tree->nodes;
    delete tree->reuse_me;
    (*parent_class->finalize) (object);
  }

  void
  pan_tree_class_init (PanTreeStoreClass *klass)
  {
    GObjectClass *object_class;

    parent_class = (GObjectClass*) g_type_class_peek_parent (klass);
    object_class = (GObjectClass*) klass;

    object_class->finalize = pan_tree_finalize;
  }

  void
  pan_tree_model_init (GtkTreeModelIface *iface)
  {
    iface->get_flags       = get_flags;
    iface->get_n_columns   = get_n_columns;
    iface->get_column_type = get_column_type;
    iface->get_iter        = get_iter;
    iface->get_path        = get_path;
    iface->get_value       = get_value;
    iface->iter_next       = iter_next;
    iface->iter_children   = iter_children;
    iface->iter_has_child  = iter_has_child;
    iface->iter_n_children = iter_n_children;
    iface->iter_nth_child  = iter_nth_child;
    iface->iter_parent     = iter_parent;
  }

  void
  pan_tree_sortable_init (GtkTreeSortableIface *iface)
  {
    iface->get_sort_column_id = get_sort_column_id;
    iface->set_sort_column_id = set_sort_column_id;
    iface->set_sort_func = set_sort_func;
    iface->set_default_sort_func = set_default_sort_func;
    iface->has_default_sort_func = has_default_sort_func;
  }

  void
  pan_tree_init (PanTreeStore * tree)
  {
    tree->stamp = g_random_int();
    tree->n_columns = 0;
    tree->nodes = new std::deque<PanTreeStore::Node>(1);
    tree->reuse_me = new std::deque<PanTreeStore::Node*>();
    tree->root = &(*tree->nodes)[0];
    tree->column_types = new std::vector<GType>();
    tree->sort_info = new std::map<int,PanTreeStore::SortInfo>();
    tree->sort_column_id = GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID;
    tree->order = GTK_SORT_ASCENDING;
  }

  PanTreeStore::Node*
  pan_tree_alloc_node (PanTreeStore * tree)
  {
    PanTreeStore::Node * node (0);

    if (!tree->reuse_me->empty()) {
      node = tree->reuse_me->back ();
      tree->reuse_me->pop_back();
    } else {
      tree->nodes->resize (tree->nodes->size() + 1);
      node = &tree->nodes->back();
    }

    g_assert (node);
    g_assert (!node->columns);
    g_assert (!node->parent);
    g_assert (node->child_index == -1);
    const int n_cols (tree->n_columns);
    node->columns = g_new0 (GValue, n_cols);
    for (int i=0; i<n_cols; ++i)
      g_value_init (&node->columns[i], (*tree->column_types)[i]);

    return node;
  }
}

/***
****
***/

GType
pan_tree_store_get_type (void)
{
  static GType pan_tree_type (0);
  if (pan_tree_type)
    return pan_tree_type;

  static const GTypeInfo pan_tree_info = {
    sizeof (PanTreeStoreClass),
    NULL, // base_init
    NULL, // base_finalize
    (GClassInitFunc) pan_tree_class_init,
    NULL, // class finalize
    NULL, // class_data
    sizeof (PanTreeStore),
    0, // n_preallocs
    (GInstanceInitFunc) pan_tree_init
  };

  pan_tree_type = g_type_register_static (G_TYPE_OBJECT, "PanTreeStore", &pan_tree_info, (GTypeFlags)0);
  static const GInterfaceInfo tree_model_info = { (GInterfaceInitFunc)pan_tree_model_init, 0, 0 };
  g_type_add_interface_static (pan_tree_type, GTK_TYPE_TREE_MODEL, &tree_model_info);
  static const GInterfaceInfo sortable_info = { (GInterfaceInitFunc)pan_tree_sortable_init, 0, 0 };
  g_type_add_interface_static (pan_tree_type, GTK_TYPE_TREE_SORTABLE, &sortable_info);
  return pan_tree_type;
}

PanTreeStore *
pan_tree_store_new (int n_columns, ...)
{
  g_return_val_if_fail (n_columns>0, 0);

  PanTreeStore * tree = (PanTreeStore*) g_object_new (PAN_TREE_STORE_TYPE, NULL);

  va_list args;
  va_start (args, n_columns);
  for (int i=0; i<n_columns; ++i) {
    const GType type = va_arg (args, GType);
    tree->column_types->push_back (type);
  }
  va_end (args);
  tree->n_columns = tree->column_types->size();

  return tree;
}

/***
****
***/

#if 0
namespace
{
  bool the_kids_are_alright (const PanTreeStore::Node * parent)
  {
    if (!parent) return false;
    if (parent->n_children != (int)parent->children.size()) return false;
    for (int i=0, n=parent->n_children; i<n; ++i) {
      const PanTreeStore::Node * child = parent->children[i];
      if (child == NULL) return false;
      if (child->parent != parent) return false;
      if (child->child_index != i) return false;
    }
    return true;
  }
}
#endif

void
pan_tree_store_insert (PanTreeStore     * tree,
                       GtkTreeIter * iter,
                       GtkTreeIter * parent,
                       gint          position)
{
  g_return_if_fail (IS_PAN_TREE_STORE(tree));
  g_return_if_fail (position >= 0);

  // insert the row
  PanTreeStore::Node * node = parent ? static_cast<PanTreeStore::Node*>(parent->user_data) : tree->root;
  PanTreeStore::Node * child = pan_tree_alloc_node (tree);
  const int old_size = node->n_children;
  //g_assert (the_kids_are_alright (node));
  if (position < old_size) {
    node->children.insert (node->children.begin()+position, child);
    ++node->n_children;
    child->parent = node;
    PanTreeStore::Node** it = &node->children[position];
    for (int i=position; i<old_size+1; )
      (*it++)->child_index = i++;
    //g_assert (the_kids_are_alright (node));
  }
  else {
    child->child_index = old_size;
    child->parent = node;
    node->children.push_back (child);
    ++node->n_children;
    //g_assert (the_kids_are_alright (node));
  }

  // set the return iter
  set_iter (iter, tree->stamp, child);

  // emit some signals
  GtkTreeModel * model = GTK_TREE_MODEL(tree);
  GtkTreePath * path = get_path (model, iter);
  //std::cerr << LINE_ID << " emitting row inserted on " << gtk_tree_path_to_string(path) << std::endl;
  gtk_tree_model_row_inserted (model, path, iter);
  if (node != tree->root) {
    gtk_tree_path_up (path);
    //std::cerr << LINE_ID << " emitting row has child toggled on " << gtk_tree_path_to_string(path) << std::endl;
    gtk_tree_model_row_has_child_toggled (model, path, parent);
  }
  gtk_tree_path_free (path);
  //g_assert (the_kids_are_alright (node));
}

void
pan_tree_store_insert_before (PanTreeStore     * tree,
                              GtkTreeIter * iter,
                              GtkTreeIter * parent,
                              GtkTreeIter * sibling)
{
  pan_tree_store_insert (tree, iter, parent, sibling ? static_cast<PanTreeStore::Node*>(sibling->user_data)->child_index : INT_MAX);
}

void
pan_tree_store_insert_after (PanTreeStore     * tree,
                             GtkTreeIter * iter,
                             GtkTreeIter * parent,
                             GtkTreeIter * sibling)
{
  pan_tree_store_insert (tree, iter, parent, sibling ? static_cast<PanTreeStore::Node*>(sibling->user_data)->child_index+1 : 0);
}

void pan_tree_store_prepend  (PanTreeStore *tree,
                              GtkTreeIter *iter,
                              GtkTreeIter *parent)
{
  pan_tree_store_insert (tree, iter, parent, 0);
}

void
pan_tree_store_append (PanTreeStore *tree,
                       GtkTreeIter *iter,
                       GtkTreeIter *parent)
{
  pan_tree_store_insert (tree, iter, parent, INT_MAX);
}

/***
****
***/

namespace
{
  bool real_set_value (PanTreeStore      * tree,
                       GtkTreeIter  * iter,
                       int            column,
                       GValue       * value,
                       gboolean       sort)
  {
    g_return_val_if_fail (IS_PAN_TREE_STORE(tree), false);
    g_return_val_if_fail (iter && iter->stamp==tree->stamp, false);
    g_return_val_if_fail (0<=column && column<tree->n_columns, false);
    g_return_val_if_fail (G_IS_VALUE(value), false);

    // if value isn't of the right type, transform it.
    bool converted (false);
    GValue real_value = {0, };
    if (!g_type_is_a (G_VALUE_TYPE (value), (*tree->column_types)[column]))
    {
      if (!(g_value_type_compatible (G_VALUE_TYPE(value), (*tree->column_types)[column]) &&
            g_value_type_compatible ((*tree->column_types)[column], G_VALUE_TYPE(value))))
      {
        g_warning ("%s: Unable to convert from %s to %s\n",
                   G_STRLOC,
                   g_type_name (G_VALUE_TYPE (value)),
                   g_type_name ((*tree->column_types)[column]));
        return false;
      }
      if (!g_value_transform (value, &real_value))
      {
        g_warning ("%s: Unable to make conversion from %s to %s\n",
                   G_STRLOC,
                   g_type_name (G_VALUE_TYPE (value)),
                   g_type_name ((*tree->column_types)[column]));
        g_value_unset (&real_value);
        return false;
      }
      converted = true;
    }

    // set the value
    PanTreeStore::Node * node = static_cast<PanTreeStore::Node*>(iter->user_data);
    if (!converted)
      g_value_copy (value, &node->columns[column]);
    else {
      g_value_copy (&real_value, &node->columns[column]);
      g_value_unset (&real_value);
    }

    if (sort && IS_SORTED(tree))
      pan_tree_store_sort_iter_changed (tree, iter, column);

    return true;
  }
}

void
pan_tree_store_set (PanTreeStore      * tree,
                    GtkTreeIter  *iter,
                    ...)
{
  va_list var_args;
  va_start (var_args, iter);
  pan_tree_store_set_valist (tree, iter, var_args);
  va_end (var_args);
}

void
pan_tree_store_set_valist (PanTreeStore      * tree,
                           GtkTreeIter  * iter,
                           va_list        var_args)
{
  g_return_if_fail (IS_PAN_TREE_STORE(tree));
  g_return_if_fail (iter);
  g_return_if_fail (iter->stamp == tree->stamp);

  if (IS_SORTED(tree))
    g_return_if_fail (has_sort_func (GTK_TREE_SORTABLE(tree), tree->sort_column_id));

  bool maybe_need_sort (false); //func != pan_tree_data_list_compare_func;
  bool emit_signal (false);
  int column = va_arg (var_args, gint);
  while (column != -1)
  {
    if (column >= tree->n_columns) {
      g_warning ("%s: Invalid column number %d added to iter (remember to end your list of columns with a -1)", G_STRLOC, column);
      break;
    }

    GValue value = { 0, };
    g_value_init (&value, (*tree->column_types)[column]);
    gchar * error = 0;
    G_VALUE_COLLECT (&value, var_args, 0, &error);
    if (error) {
      g_warning ("%s: %s", G_STRLOC, error);
      g_free (error);
      /* we purposely leak the value here, it might not be
       * in a sane state if an error condition occured */
      break;
    }

    emit_signal = real_set_value (tree, iter, column, &value, false) || emit_signal;
    g_value_unset (&value);

    maybe_need_sort |= column == tree->sort_column_id;

    column = va_arg (var_args, gint);
  }

  if (maybe_need_sort && IS_SORTED(tree))
    pan_tree_store_sort_iter_changed (tree, iter, tree->sort_column_id);

  if (emit_signal) {
    // std::cerr << LINE_ID << " emitting row_changed signal " << std::endl;
    GtkTreePath * path = gtk_tree_model_get_path (GTK_TREE_MODEL(tree), iter);
    gtk_tree_model_row_changed (GTK_TREE_MODEL(tree), path, iter);
    gtk_tree_path_free (path);
  }
}

/***
****
***/

namespace
{
  struct NodeWalker
  {
    virtual ~NodeWalker() {}
    virtual void operator()(PanTreeStore::Node* node) = 0;
  };

  void node_walk (NodeWalker& func, PanTreeStore::Node * node)
  {
    for (int i=0, n=node->n_children; i<n; ++i)
      node_walk (func, node->children[i]);
    func (node);
  }

  struct UnsetNodeWalker: public NodeWalker
  {
    PanTreeStore * tree;
    UnsetNodeWalker(PanTreeStore* t): tree(t) {}
    virtual ~UnsetNodeWalker() {}
    virtual void operator()(PanTreeStore::Node* node) {
      pan_tree_node_unset_values (tree, node);
      node->parent = 0;
      node->child_index = -1;
      //tree->reuse_me->push_back (node);
    }
  };
}

gboolean
pan_tree_store_remove (PanTreeStore *tree, GtkTreeIter *iter)
{
  g_return_val_if_fail (IS_PAN_TREE_STORE(tree), false);
  g_return_val_if_fail (iter, false);
  g_return_val_if_fail (iter->stamp == tree->stamp, false);

  GtkTreeModel * model = GTK_TREE_MODEL (tree);
  GtkTreePath * path = gtk_tree_model_get_path (model, iter);

  // check the pre-removal state
  PanTreeStore::Node * node = static_cast<PanTreeStore::Node*>(iter->user_data);
  PanTreeStore::Node * parent = node->parent;
  //g_assert (the_kids_are_alright (parent));

  // unthread the node
  const int pos = node->child_index;
  parent->children.erase (parent->children.begin() + pos);
  --parent->n_children;
  if (pos < parent->n_children) {
    PanTreeStore::Node ** it (&parent->children[pos]);
    for (int i=pos; i<parent->n_children; ++i, ++it)
      (*it)->child_index = i;
  }
  //g_assert (the_kids_are_alright (parent));

  // revalidate the iterator
  const bool iter_is_valid_now (pos < parent->n_children);
  if (iter_is_valid_now)
    set_iter (iter, tree->stamp, parent->children[pos]);
  else
    invalidate_iter (iter);

  // let the world know the row was removed
  gtk_tree_model_row_deleted (model, path);

  // if the parent's has-children state has changed, let the world know that too
  if (parent!=tree->root && !parent->n_children) {
    GtkTreeIter pit;
    set_iter (&pit, tree->stamp, parent);
    gtk_tree_path_up (path);
    gtk_tree_model_row_has_child_toggled (model, path, &pit);
  }

  gtk_tree_path_free (path);

  // add `node' and all its descendants back to the reuse-me pool
  UnsetNodeWalker walker (tree);
  node_walk (walker, node);

  //g_assert (the_kids_are_alright (parent));
  return iter_is_valid_now;
}
